/*
 * xmms-mad - mp3 plugin for xmms
 * Copyright (C) 2001-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Much of the code in this file was originally taken from the 
 * mpg123 plugin that is part of the xmms source and is 
 * Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, 
 * Thomas Nilsson and 4Front Technologies
 * 
 * $superduper: fileinfo.c,v 1.12 2002/07/26 17:14:12 samc Exp $
 */
#include "config.h"

#include "xmms-mad.h"
#include "input.h"

#include <xmms/util.h>
#include <gtk/gtk.h>

/* #define DEBUG */

static GtkWidget *window = 0;
static GtkWidget *filename_entry, *id3_frame;
static GtkWidget *title_entry, *artist_entry, *album_entry;
static GtkWidget *year_entry, *tracknum_entry, *comment_entry;
static GtkWidget *genre_combo;
static GtkWidget *mpeg_level, *mpeg_bitrate, *mpeg_samplerate, *mpeg_flags;
static GtkWidget *mpeg_fileinfo;

static GList *genre_list = 0;
static struct mad_info_t info;
struct id3_frame * id3_frame_new (const char* str);

void
update_id3_frame (struct id3_tag *tag, const char *name, const char *data)
{
  struct id3_frame *frame;
  id3_ucs4_t *ucs4;
  int len = id3_latin1_length ((id3_latin1_t *) data);

  printf ("%s: %s\n", name, data);
  frame = id3_tag_findframe (tag, name, 0);
  ucs4 = (id3_ucs4_t *) g_malloc ((len + 1) * sizeof (*ucs4));
  id3_latin1_decode ((id3_latin1_t *) data, ucs4);

  if (!frame)
    {
      printf ("not found\n");
      frame = id3_frame_new (name);
      id3_field_setstrings (&frame->fields[1], 1, &ucs4);
      id3_tag_attachframe (tag, frame); 
    }
  else
    {
      id3_field_setstrings (&frame->fields[1], 1, &ucs4);
    }
  g_free (ucs4);
}

static gint 
genre_comp_func (gconstpointer a, gconstpointer b)
{
  return strcasecmp (a, b);
}

static void 
close_window (GtkWidget *w, gpointer data)
{
  input_term (&info);
  gtk_widget_destroy (window);
}

static void 
save_cb (GtkWidget *w, gpointer data)
{
  gchar *text;
  struct id3_file *id3file;
  struct id3_tag *id3tag;

  if (info.remote)
    return;

  /* read tag from file */
  id3file = id3_file_open (info.filename, ID3_FILE_MODE_READWRITE);
  if (!id3file)
    {
      printf ("making new tag\n");
      id3tag = id3_tag_new ();
      id3_tag_clearframes (id3tag);
      id3tag->options |= ID3_TAG_OPTION_ID3V1;
      /* id3file = id3_file_new (info.filename, id3tag);*/
      if (!id3file)
        {
          xmms_show_message ("File Info", "Couldn't open file!", "Ok", FALSE, NULL, NULL);
          return;
        }
    }

  id3tag = id3_file_tag (id3file);

  text = gtk_entry_get_text (GTK_ENTRY (title_entry));
  update_id3_frame (id3tag, "TIT2", text);

  text = gtk_entry_get_text (GTK_ENTRY (artist_entry));
  update_id3_frame (id3tag, "TPE1", text);

  text = gtk_entry_get_text (GTK_ENTRY (album_entry));
  update_id3_frame (id3tag, "TALB", text);

  text = gtk_entry_get_text (GTK_ENTRY (year_entry));
  update_id3_frame (id3tag, "TDRC", text);

  /*
  text = gtk_entry_get_text (GTK_ENTRY (comment_entry));
  id3_tag_attachframe (info.tag, make_id3_frame (text));
  */

  text = gtk_entry_get_text (GTK_ENTRY (tracknum_entry));
  update_id3_frame (id3tag, "TRCK", text);

  /*
  text = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (genre_combo)->entry));
  */

  if (id3_file_update (id3file))
    {
      xmms_show_message ("File Info", "Couldn't write tag!", "Ok", FALSE, NULL, NULL);
    }
  id3_file_close (id3file);
  gtk_widget_destroy (window);
}

static void 
remove_id3_cb (GtkWidget *w, gpointer data)
{
  xmms_show_message ("File Info", "No tag to remove!", "Ok", FALSE, NULL, NULL);
  gtk_widget_destroy (window);
}

void
create_window ()
{
  GtkWidget *vbox, *hbox, *left_vbox, *table;
  GtkWidget *mpeg_frame, *mpeg_box;
  GtkWidget *label, *filename_hbox;
  GtkWidget *bbox, *save, *remove_id3, *cancel;

  window = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
  gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(close_window), &window);
  gtk_container_set_border_width(GTK_CONTAINER(window), 10);

  vbox = gtk_vbox_new(FALSE, 10);
  gtk_container_add(GTK_CONTAINER(window), vbox);

  filename_hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), filename_hbox, FALSE, TRUE, 0);

  label = gtk_label_new("Filename:");
  gtk_box_pack_start(GTK_BOX(filename_hbox), label, FALSE, TRUE, 0);
  filename_entry = gtk_entry_new();
  gtk_editable_set_editable(GTK_EDITABLE(filename_entry), FALSE);
  gtk_box_pack_start(GTK_BOX(filename_hbox), filename_entry, TRUE, TRUE, 0);

  hbox = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);

  left_vbox = gtk_vbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(hbox), left_vbox, FALSE, FALSE, 0);

  id3_frame = gtk_frame_new("ID3 Tag:");
  gtk_box_pack_start(GTK_BOX(left_vbox), id3_frame, FALSE, FALSE, 0);

  table = gtk_table_new(5, 5, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 5);
  gtk_container_add(GTK_CONTAINER(id3_frame), table);

  label = gtk_label_new("Title:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 5, 5);

  title_entry = gtk_entry_new ();
  gtk_table_attach(GTK_TABLE(table), title_entry, 1, 4, 0, 1, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  label = gtk_label_new("Artist:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 5);

  artist_entry = gtk_entry_new ();
  gtk_table_attach(GTK_TABLE(table), artist_entry, 1, 4, 1, 2, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  label = gtk_label_new("Album:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 5, 5);

  album_entry = gtk_entry_new ();
  gtk_table_attach(GTK_TABLE(table), album_entry, 1, 4, 2, 3, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  label = gtk_label_new("Comment:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 5, 5);

  comment_entry = gtk_entry_new ();
  gtk_table_attach(GTK_TABLE(table), comment_entry, 1, 4, 3, 4, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  label = gtk_label_new("Year:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 4, 5, GTK_FILL, GTK_FILL, 5, 5);

  year_entry = gtk_entry_new ();
  gtk_widget_set_usize(year_entry, 40, -1);
  gtk_table_attach(GTK_TABLE(table), year_entry, 1, 2, 4, 5, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  label = gtk_label_new("Track number:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 2, 3, 4, 5, GTK_FILL, GTK_FILL, 5, 5);

  tracknum_entry = gtk_entry_new ();
  gtk_widget_set_usize(tracknum_entry, 40, -1);
  gtk_table_attach(GTK_TABLE(table), tracknum_entry, 3, 4, 4, 5, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  label = gtk_label_new("Genre:");
  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 5, 6, GTK_FILL, GTK_FILL, 5, 5);

  genre_combo = gtk_combo_new();
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(genre_combo)->entry), FALSE);
  if (!genre_list)
    {
      genre_list = g_list_prepend(genre_list, "");
      genre_list = g_list_sort(genre_list, genre_comp_func);
    }
  gtk_combo_set_popdown_strings(GTK_COMBO(genre_combo), genre_list);

  gtk_table_attach(GTK_TABLE(table), genre_combo, 1, 4, 5, 6, GTK_FILL | GTK_EXPAND | GTK_SHRINK, GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 5);

  bbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 5);
  gtk_box_pack_start(GTK_BOX(left_vbox), bbox, FALSE, FALSE, 0);

  save = gtk_button_new_with_label("Save");
  gtk_signal_connect(GTK_OBJECT(save), "clicked", GTK_SIGNAL_FUNC(save_cb), NULL);
  GTK_WIDGET_SET_FLAGS(save, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(bbox), save, TRUE, TRUE, 0);
  gtk_widget_grab_default(save);

  remove_id3 = gtk_button_new_with_label("Remove ID3");
  gtk_signal_connect(GTK_OBJECT(remove_id3), "clicked", GTK_SIGNAL_FUNC(remove_id3_cb), NULL);
  GTK_WIDGET_SET_FLAGS(remove_id3, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(bbox), remove_id3, TRUE, TRUE, 0);

  cancel = gtk_button_new_with_label("Cancel");
  gtk_signal_connect_object(GTK_OBJECT(cancel), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(window));
  GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
  gtk_box_pack_start(GTK_BOX(bbox), cancel, TRUE, TRUE, 0);

  mpeg_frame = gtk_frame_new("MPEG Info:");
  gtk_box_pack_start(GTK_BOX(hbox), mpeg_frame, FALSE, FALSE, 0);

  mpeg_box = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(mpeg_frame), mpeg_box);
  gtk_container_set_border_width(GTK_CONTAINER(mpeg_box), 10);
  gtk_box_set_spacing(GTK_BOX(mpeg_box), 0);

  mpeg_level = gtk_label_new("");
  gtk_widget_set_usize(mpeg_level, 120, -2);
  gtk_misc_set_alignment(GTK_MISC(mpeg_level), 0, 0);
  gtk_box_pack_start(GTK_BOX(mpeg_box), mpeg_level, FALSE, FALSE, 0);

  mpeg_bitrate = gtk_label_new("");
  gtk_misc_set_alignment(GTK_MISC(mpeg_bitrate), 0, 0);
  gtk_label_set_justify(GTK_LABEL(mpeg_bitrate), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(mpeg_box), mpeg_bitrate, FALSE, FALSE, 0);

  mpeg_samplerate = gtk_label_new("");
  gtk_misc_set_alignment(GTK_MISC(mpeg_samplerate), 0, 0);
  gtk_box_pack_start(GTK_BOX(mpeg_box), mpeg_samplerate, FALSE, FALSE, 0);

  mpeg_flags = gtk_label_new("");
  gtk_misc_set_alignment(GTK_MISC(mpeg_flags), 0, 0);
  gtk_label_set_justify(GTK_LABEL(mpeg_flags), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(mpeg_box), mpeg_flags, FALSE, FALSE, 0);

  mpeg_fileinfo = gtk_label_new("");
  gtk_misc_set_alignment(GTK_MISC(mpeg_fileinfo), 0, 0);
  gtk_label_set_justify(GTK_LABEL(mpeg_fileinfo), GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(mpeg_box), mpeg_fileinfo, FALSE, FALSE, 0);

  gtk_widget_show_all(window);
}

void 
id3_frame_to_entry (char *framename, GtkEntry *entry)
{
  gchar *text;
  struct id3_frame const *frame;
  union id3_field const *field;
  gtk_entry_set_text (entry, "");
  if (!info.tag)
    {
      return;
    }
  if (strcmp(framename, "COMM") == 0)
    {
      int i = 0;
      id3_ucs4_t const* ucs4;
      while ((frame = id3_tag_findframe (info.tag, framename, i++)))
        {
          ucs4 = id3_field_getstring(&frame->fields[2]);
          if (!*ucs4)
            {
              ucs4 = id3_field_getfullstring(&frame->fields[3]);
              text = id3_ucs4_latin1duplicate (ucs4);
              gtk_entry_set_text (entry, text);
              g_free (text);
            }
        }
    }
  else
    {
      frame = id3_tag_findframe (info.tag, framename, 0);
      if (frame)
        {
          field = &frame->fields[1];
          text = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
          gtk_entry_set_text (entry, text);
          g_free (text);
        }
    }
}

void 
xmmsmad_get_file_info (char *filename)
{
  gchar *title;
  const gchar *emphasis[4];
  const gchar *bool_label[2];

#ifdef DEBUG
  printf ("f: xmmsmad_get_file_info: %s\n", filename);
#endif

  emphasis[0] = "None";
  emphasis[1] = "50/15 ms";
  emphasis[2] = "";
  emphasis[3] = "CCIT J.17";
  bool_label[0] = "No";
  bool_label[1] = "Yes";

  create_window ();

  input_init (&info, filename);
  input_get_info (&info);

  title = g_strdup_printf("File Info - %s", g_basename(filename));
  gtk_window_set_title(GTK_WINDOW(window), title);
  g_free(title);

  gtk_entry_set_text(GTK_ENTRY(filename_entry), filename);
  gtk_editable_set_position(GTK_EDITABLE(filename_entry), -1);

  id3_frame_to_entry ("TPE1", GTK_ENTRY (artist_entry));
  id3_frame_to_entry ("TIT2", GTK_ENTRY (title_entry));
  id3_frame_to_entry ("TALB", GTK_ENTRY (album_entry));
  id3_frame_to_entry ("TDRC", GTK_ENTRY (year_entry));
  id3_frame_to_entry ("TRCK", GTK_ENTRY (tracknum_entry));
  id3_frame_to_entry ("COMM", GTK_ENTRY (comment_entry));
  gtk_label_set_text (GTK_LABEL (mpeg_level), "MPEG ?, layer ?");
  gtk_label_set_text (GTK_LABEL (mpeg_bitrate), "");
  gtk_label_set_text (GTK_LABEL (mpeg_samplerate), "");
  gtk_label_set_text (GTK_LABEL (mpeg_flags), "");
  gtk_label_set_text (GTK_LABEL (mpeg_fileinfo), "");
  gtk_list_select_item (GTK_LIST (GTK_COMBO(genre_combo)->list), g_list_index(genre_list, ""));

  gtk_widget_set_sensitive (id3_frame, TRUE);

  gtk_list_select_item (GTK_LIST (GTK_COMBO (genre_combo)->list), g_list_index (genre_list, (char *) id3_genre_name ((id3_ucs4_t *) "1")));
}

