/*
 * xmms-mad - mp3 plugin for xmms
 * Copyright (C) 2001-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $superduper: input.c,v 1.28 2002/08/22 22:57:31 samc Exp $
 */
#include "config.h"

#ifdef HAVE_ASSERT_H
#include <assert.h>
#endif /* HAVE_ASSERT_H */

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */

#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif /* HAVE_ARPA_INET_H */

#ifdef HAVE_NETDB_H 
#include <netdb.h>
#endif /* HAVE_NETDB_H */

#ifdef HAVE_SYS_STAT_H 
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */

#ifdef HAVE_SYS_TIME_H 
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */

#include <fcntl.h>
#include <errno.h>

#include "input.h"

#define DIR_SEPARATOR '/'
#define HEADER_SIZE 256
#define LINE_LENGTH 256

/*#define DEBUG 1*/

int do_scan_file = 1;

extern void scan_file (struct mad_info_t *info);

struct connection_t
{
  struct hostent *hostinfo;
  struct sockaddr_in address;
  int sock;
  int udpsock;
  int udpport;
};

struct streamdata_t
{
  char *f_buffer;
  char *b_buffer;
  int f_buffer_start;
  int f_buffer_end;
  int b_buffer_end;
  char *name;
  char *notice1;
  char *notice2;
  char *genre;
  char *url;
  int metaint;
  int bitrate;
};

static void
input_parse_url (struct mad_info_t *info)
{
  char *file;
  char *port;
  char *host = info->url;

  info->remote = 0;
  if (!strncasecmp ("http://", host, 7))
    {
      info->remote = 1;
      /* skip leading "http://" */
      host += 7;
    }
  else
    {
      info->filename = info->url;
      return;
    }

  /* parse host:port/filename */
  if ((port = strchr (host, ':')))
    {
      *port = '\0';
      port++;
      if ((file = strchr (port, '/')))
        {
          info->filename = strdup (file);
          *file = '\0';
          file++;
        }
      info->port = atoi (port);
    }
  else
    {
      if ((file = strchr (host, '/')))
        {
          *file = '\0';
          file++;
          info->filename = strdup (file);
        }
    }
  info->host = strdup (host);
  if (!info->filename)
    {
      info->filename = strdup ("/");
    }
}

/**
 * Read a line from the network socket.
 */
int
input_http_readline (struct mad_info_t *madinfo, char *buffer,
                     int buffer_size)
{
  int i;
  assert (madinfo && madinfo->connection);
  for (i = 0; i < buffer_size; i++)
    {
      read (madinfo->connection->sock, buffer + i, 1);
      if (buffer[i] == '\n')
        break;
      if (buffer[i] == '\r')
        i--;
    }
  buffer[i] = '\0';
  return i;
}

/**
 * Open the UDP title streaming socket.
 */
int
input_udp_init (struct mad_info_t *madinfo)
{
  struct sockaddr_in sin;
  int sinlen = sizeof (struct sockaddr_in);
  struct connection_t *con;

  con = madinfo->connection;
  con->udpsock = socket (AF_INET, SOCK_DGRAM, 0);

  if (con->udpsock == -1)
    {
      xmmsmad_error ("failed to init UDP socket");
      return -1;
    }

  memset (&sin, 0, sinlen);
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htonl (INADDR_ANY);

  if (bind (con->udpsock, (struct sockaddr *) &sin, sinlen))
    {
      xmmsmad_error ("failed to bind to a UDP port");
      return -1;
    }
  fcntl (con->udpsock, F_SETFL, O_NONBLOCK);
  memset (&sin, 0, sinlen);
  if (getsockname (con->udpsock, (struct sockaddr *) &sin, &sinlen) < 0)
    {
      xmmsmad_error ("failed to retrieve socket info: %s", strerror (errno));
      close (con->udpsock);
      return -1;
    }

  con->udpport = ntohs (sin.sin_port);
  /* printf ("Listening on local %s:%d\n", inet_ntoa (sin.sin_addr), con->udpport); */

  return 0;
}

/**
 * Update title from UDP socket
 */
int
input_udp_read (struct mad_info_t *info)
{
  char buffer[LINE_LENGTH];
  int len;
  assert (info && info->connection);
  len = recv (info->connection->udpsock, buffer, LINE_LENGTH, 0);
  /* read (madinfo->connection->udpsock, buffer, LINE_LENGTH); */
  if (len > 0)
    {
      printf ("Got some UDP:\n");
      buffer[len] = '\0';
      printf ("%s\n", buffer);
    }
  return 0;
}

/**
 * Open the http connection and the UDP title stream connection
 */
static int
input_http_init (struct mad_info_t *info)
{
  char header[HEADER_SIZE];
  char buffer[LINE_LENGTH];
  struct connection_t *con;

  if (!info->sdata)
    {
      info->sdata = (struct streamdata_t *)
        g_malloc (sizeof (struct streamdata_t));
    }

  if (!info->connection)
    {
      info->connection = (struct connection_t *)
        g_malloc (sizeof (struct connection_t));
    }

  /* reset data */
  memset (info->sdata, 0, sizeof (struct streamdata_t));
  memset (info->connection, 0, sizeof (struct connection_t));

  /* malloc http buffer */
  info->sdata->b_buffer = (char *) g_malloc (HTTP_BUFFER_SIZE * sizeof (char));
  info->sdata->f_buffer = (char *) g_malloc (HTTP_BUFFER_SIZE * sizeof (char));

  con = info->connection;

  con->sock = socket (AF_INET, SOCK_STREAM, 0);
  con->address.sin_family = AF_INET;

  if (!(con->hostinfo = gethostbyname (info->host)))
    {
      xmmsmad_error ("failed to look up host: %s", info->host);
      return -1;
    }

  memcpy (&con->address.sin_addr.s_addr,
          *(con->hostinfo->h_addr_list),
          sizeof (con->address.sin_addr.s_addr));
  con->address.sin_port = htons (info->port);

  if (connect (con->sock, (struct sockaddr *)
               &con->address, sizeof (struct sockaddr_in)) == -1)
    {
      if (errno != EINPROGRESS)
        {
          xmmsmad_error ("couldn't connect to host %s", info->host);
          return -1;
        }
    }
  snprintf (header, HEADER_SIZE - 1, "GET %s HTTP/1.0\r\n"
            "Host: %s\r\n"
            "User-Agent: %s/%s\r\n"
            "x-audiocast-udpport: %d\r\n"
            "\r\n", info->filename, info->host, PACKAGE, "0.1", con->udpport);

  /* "Icy-MetaData:1\r\n" */
  write (con->sock, header, strlen (header));

  input_http_readline (info, buffer, LINE_LENGTH);
  if (strncmp (buffer, "ICY 200", 7))
    {
      xmmsmad_error ("server said: %s", buffer);
      return -1;
    }

  while (input_http_readline (info, buffer, LINE_LENGTH) > 0)
    {
      if (!strncmp (buffer, "icy-name:", 9))
        info->sdata->name = strdup (buffer + 9);
      else if (!strncmp (buffer, "x-audiocast-name:", 17))
        info->sdata->name = strdup (buffer + 17);
      else if (!strncmp (buffer, "icy-metaint:", 12))
        info->sdata->metaint = atoi (buffer + 12);
      else if (!strncmp (buffer, "icy-url:", 8))
        info->sdata->url = strdup (buffer + 8);
      else if (!strncmp (buffer, "icy-genre:", 10))
        info->sdata->genre = strdup (buffer + 10);
      else if (!strncmp (buffer, "icy-notice1:", 12))
        info->sdata->notice1 = strdup (buffer + 12);
      else if (!strncmp (buffer, "icy-notice2:", 12))
        info->sdata->notice2 = strdup (buffer + 12);
      else if (!strncmp (buffer, "icy-br:", 7))
        info->sdata->bitrate = atoi (buffer + 7);
      else if (!strncmp (buffer, "icy-pub:", 8))
        {
        }                       /* nothing */
      else
        printf ("unknown header: %s\n", buffer);
    }

  info->title = strdup (info->sdata->name);
  info->bitrate = info->sdata->bitrate;
  info->length = -1;

  fcntl (con->sock, F_SETFL, O_NONBLOCK);
  return 0;
}

/**
 * init the mad_info_t struct.
 */
int
input_init (struct mad_info_t *info, const char *url)
{
  char *pos;
#ifdef DEBUG
  printf ("f: input_init\n");
#endif
  info->fmt = FMT_S16_LE;
  info->channels = 2;
  info->size = -1;
  info->freq = 44100;
  info->port = 80;
  info->length = -1;
  info->pos = mad_timer_zero;
  info->url = strdup (url);
  info->currentframe = 0;
  info->framecount = 0;
  info->bitrate = 0;

  input_parse_url (info);
  pos = strrchr (info->filename, DIR_SEPARATOR);
  if (pos)
    {
      info->title = strdup (pos + 1);
    }
  else
    {
      info->title = strdup (info->filename);
    }
  if (!info->remote)
    {
      struct stat buf;
      info->infile = open (info->filename, O_RDONLY);
      if (info->infile == -1)
        {
          /* can't show error here.. why? */
          /* xmmsmad_error ("failed to open: %s", info->filename); */
          return -1;
        }
      if (fstat (info->infile, &buf) == -1)
        {
          /* can't show error here.. why? */
          /* xmmsmad_error ("failed to stat: %s", info->filename); */
          return -1;
        }
      info->size = buf.st_size;
    }
#ifdef DEBUG
  printf ("f: input_init done\n");
#endif
  return 0;
}

/**
 * Retrieve meta-information about URL.
 * For local files this means ID3 tag etc.
 */
int
input_get_info (struct mad_info_t *info)
{
#ifdef DEBUG
  printf ("f: input_get_info: %s\n", info->title);
#endif
  if (info->remote)
    {
      if (input_http_init (info))
        return -1;
      if (input_udp_init (info))
        return -1;
    }
  else
    {
      /* local mp3 file */
      struct id3_frame const *frame;
      union id3_field const *field;
      TitleInput *title_input = (TitleInput *) g_malloc (sizeof (TitleInput));
      XMMS_NEW_TITLEINPUT (title_input);

      /* read the ID3 tag */
      info->id3file = id3_file_open (info->filename, ID3_FILE_MODE_READONLY);
      if (!info->id3file)
        {
          return 0;
        }
      info->tag = id3_file_tag (info->id3file);
      if (info->tag)
        {
          frame = id3_tag_findframe (info->tag, ID3_FRAME_ARTIST, 0);
          if (frame && (field = &frame->fields[1]))
            {
              title_input->performer = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
            }
          frame = id3_tag_findframe (info->tag, ID3_FRAME_TITLE, 0);
          if (frame && (field = &frame->fields[1]))
            {
              title_input->track_name = id3_ucs4_latin1duplicate (id3_field_getstrings (field, 0));
            }
        }
      /* scan the mp3 file */
      if (do_scan_file)
          scan_file (info);
      lseek (info->infile, 0, SEEK_SET);
      info->title = xmms_get_titlestring (xmms_get_gentitle_format (), title_input);
      if (!info->title) 
        info->title = strdup (info->filename);
      if (title_input->track_name)
        g_free (title_input->track_name);
      if (title_input->performer)
        g_free (title_input->performer);
      g_free (title_input);
    }
#ifdef DEBUG
  printf ("f: input_get_info done\n");
#endif
  return 0;
}

int
read_from_socket (struct mad_info_t *madinfo)
{
  static int bytes_read = 0;
  int free_space;
  int len;
  assert (madinfo && madinfo->connection);
  free_space = HTTP_BUFFER_SIZE - madinfo->sdata->b_buffer_end;
  input_udp_read (madinfo);
  len = read (madinfo->connection->sock,
              madinfo->sdata->b_buffer + madinfo->sdata->b_buffer_end,
              free_space);
  if (len > 0)
    madinfo->sdata->b_buffer_end += len;
  bytes_read += len;
/*
  if (bytes_read > sdata.metaint)
    {
      char * metadata = sdata.b_buffer+sdata.b_buffer_end - ( bytes_read + len - sdata.metaint );
      printf ("numblocks = %d\n", metadata[0]);
      bytes_read = bytes_read + len - sdata.metaint;
    }
*/
  return len;
}

int
input_rebuffer (struct mad_info_t *madinfo)
{
  int total = 0;
  int data_read = 0;
  int len = 0;
  fd_set rfds;
  struct timeval tv;
  struct connection_t *con;

  assert (madinfo && madinfo->connection);
  con = madinfo->connection;
  total = HTTP_BUFFER_SIZE - madinfo->sdata->b_buffer_end;
  /* Wait up up to thirty seconds. */
  tv.tv_sec = 30;
  tv.tv_usec = 0;

  mad_plugin.set_info ("BUFFERING", -1, 0, 0, madinfo->channels);
  FD_ZERO (&rfds);
  FD_SET (con->sock, &rfds);
  while (data_read < total)
    {
      int rtn = select (con->sock + 1, &rfds, 0, 0, &tv);
      if (rtn == 0)
        {
          xmmsmad_error ("timeout while waiting for data from server");
          return -1;
        }
      if (rtn < 0)
        {
          xmmsmad_error ("error reading from socket");
          return -1;
        }
      len = read (con->sock,
                  madinfo->sdata->b_buffer + madinfo->sdata->b_buffer_end,
                  total - data_read);
      if (len == -1)
        return -1;
      madinfo->sdata->b_buffer_end += len;
      data_read += len;
    }

  /* make sure the back buffer is now full */
  assert (madinfo->sdata->b_buffer_end == HTTP_BUFFER_SIZE);
  return 0;
}

/**
 * Read data from the source given my madinfo into the buffer
 * provided.  Return the number of bytes read.
 * @return 0 on EOF
 * @return -1 on error
 */
int
input_get_data (struct mad_info_t *madinfo, char *buffer, int buffer_size)
{
  int len = 0;
  int remainder = 0;
#ifdef DEBUG
  printf ("f: input_get_data: %d\n", buffer_size);
#endif
  if (madinfo->remote)
    {
      int available;
      struct streamdata_t *sdata = madinfo->sdata;
      read_from_socket (madinfo);
      available = +sdata->b_buffer_end
        + sdata->f_buffer_end - sdata->f_buffer_start;
      /* make sure we are not asking for too much at one time */
      assert (buffer_size < HTTP_BUFFER_SIZE);
      if (available < buffer_size)
        {
          /* we ran out completely rebuffer */
          if (input_rebuffer (madinfo) == -1)
            {
              xmmsmad_error ("error filling http buffer");
            }
        }

      if (buffer_size > sdata->f_buffer_end - sdata->f_buffer_start)
        {
          char *tmp;

          /* copy out the remainder of the front buffer */
          len = sdata->f_buffer_end - sdata->f_buffer_start;
          memcpy (buffer, sdata->f_buffer + sdata->f_buffer_start, len);

          /* swap the buffers */
          /* printf ("swapping buffers\n"); */
          tmp = sdata->f_buffer;
          sdata->f_buffer = sdata->b_buffer;
          sdata->f_buffer_start = 0;
          sdata->f_buffer_end = sdata->b_buffer_end;

          sdata->b_buffer = tmp;
          sdata->b_buffer_end = 0;
        }

      remainder = buffer_size - len;
      if (remainder)
        {
          assert (remainder < sdata->f_buffer_end - sdata->f_buffer_start);
          memcpy (buffer + len, sdata->f_buffer + sdata->f_buffer_start,
                  remainder);
          sdata->f_buffer_start += remainder;
          len += remainder;
        }
    }
  else
    {
      /* simply read to data from the file */
      assert (madinfo->infile >= 0);
      len = read (madinfo->infile, buffer, buffer_size);
    }
#ifdef DEBUG
  printf ("f: input_get_data done: %d\n", len);
#endif
  return len;
}

/**
 * Free up all input related memory.
 */
int
input_term (struct mad_info_t *info)
{
#ifdef DEBUG
  printf ("f: input_term\n");
#endif
  if (info->id3file)
    {
      id3_file_close (info->id3file);
      info->id3file = 0;
    }
  if (info->title)
    g_free (info->title);
  if (info->url)
    g_free (info->url);
  if (info->connection)
    g_free (info->connection);
  if (info->infile)
    close (info->infile);
  if (info->sdata)
    {
      g_free (info->sdata->b_buffer);
      g_free (info->sdata->f_buffer);
      g_free (info->sdata);
    }
#ifdef DEBUG
  printf ("f: input_term done\n");
#endif
  return 0;
}
