/*
 * xmms-mad - mp3 plugin for xmms
 * Copyright (C) 2001-2002 Sam Clegg - See COPYING
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $superduper: xmms-mad.c,v 1.29 2002/08/22 23:06:54 samc Exp $
 */
#include "config.h"
#include "xmms-mad.h"
#include "input.h"

#include <pthread.h>
#include <gtk/gtk.h>
#include <xmms/util.h>
#include <stdarg.h>
#include <fcntl.h>


/*#define DEBUG*/

static pthread_t decode_thread;
struct mad_info_t info;
InputPlugin mad_plugin;

static GtkWidget *about_window = 0;
static GtkWidget *error_dialog = 0;

static void
xmmsmad_init ()
{
  memset (&info, 0, sizeof (struct mad_info_t));
}

static void
xmmsmad_cleanup ()
{
}

static void
xmmsmad_configure ()
{
}

static void
xmmsmad_set_eq (int on, float preamp, float *bands)
{
}

static int
xmmsmad_is_our_file (char *filename)
{
  int fin;
  unsigned long check = 0L;

  if (strncasecmp ("http", filename, 4) == 0)
    {
      /* [FIXME] do some checks here to make sure we have an mp3 stream */
      return 1;
    }

  fin = open (filename, O_RDONLY);
  if (fin >= 0 && read (fin, &check, 4) == 4)
    {
      /* If first two bytes are a sync header or three bytes are "ID3" */
      if ( ((check & 0xf0ff) == 0xf0ff) || ((check & 0xffffff) == 0x334449)) 
        {
          close (fin);
          return 1;
        }
      if (check == 0x46464952) /* RIFF RMP3 stream */
        {	
          lseek (fin, 4, SEEK_CUR);
          if (read (fin, &check, 4) == 4 && check == 0x33504d52) 
            {
              close (fin);
              return 1;
            }
        }
    }

  close (fin);
  return 0;
}

static void
xmmsmad_stop ()
{
#ifdef DEBUG
  g_message ("f: xmmsmad_stop");
#endif /* DEBUG */
  if (info.playing)
    {
      info.stop = 1;
#ifdef DEBUG
      g_message ("waiting for thread");
#endif /* DEBUG */
/*      pthread_cancel (decode_thread);*/
      pthread_join (decode_thread, 0);
#ifdef DEBUG
      g_message ("thread done");
#endif /* DEBUG */
    }
}

static void
xmmsmad_play_file (char *url)
{
  xmmsmad_stop ();

#ifdef DEBUG
  g_message ("playing %s", url);
#endif /* DEBUG */
  info.playing = 1;
  if (!input_init (&info, url))
    {
      input_get_info (&info);
      pthread_create (&decode_thread, 0, decode, (void *) &info);
    }
  else
    info.playing = 0;
}

static void
xmmsmad_pause (short paused)
{
  mad_plugin.output->pause (paused);
}

static void
xmmsmad_seek (int time)
{
  /* xmms gives us the desired seek time in seconds */
  info.seek = time * 1000;
}

static int
xmmsmad_get_time ()
{
  if (!info.playing)
    {
#ifdef DEBUG
      g_message ("returning -1 from get_time");
#endif
      return -1;
    }
  return mad_plugin.output->output_time ();
}

/**
 * Scan the file or URL given.
 */
static void
xmmsmad_get_song_info (char *url, char **title, int *length)
{
  struct mad_info_t myinfo;
#ifdef DEBUG
  g_message ("f: xmmsmad_get_song_info: %s", url);
#endif
  memset (&myinfo, 0, sizeof (struct mad_info_t));
  input_init (&myinfo, url);
  if (!input_get_info (&myinfo))
    {
      *title  = strdup (myinfo.title);
      *length = myinfo.length;
    }
  else
    {
      *title = strdup (url);
      *length = -1;
    }
  input_term (&myinfo);
}

/**
 * Pop-up GTK about box.
 * Taken from mpg123 plugin.
 */
static void
xmmsmad_about ()
{
  GtkWidget *dialog_vbox1;
  GtkWidget *hbox1;
  GtkWidget *label1;
  GtkWidget *dialog_action_area1;
  GtkWidget *about_exit;

  if (!about_window)
    {
      about_window = gtk_dialog_new ();
      gtk_object_set_data (GTK_OBJECT (about_window), "about_window", about_window);
      gtk_window_set_title (GTK_WINDOW (about_window), "About MPEG Layer 1/2/3 plugin (support 3DNow!)");
      gtk_window_set_policy (GTK_WINDOW (about_window), FALSE, FALSE, FALSE);
      gtk_window_set_position (GTK_WINDOW (about_window), GTK_WIN_POS_MOUSE);
      gtk_signal_connect (GTK_OBJECT (about_window), "destroy", GTK_SIGNAL_FUNC (gtk_widget_destroyed), &about_window);
      gtk_container_border_width (GTK_CONTAINER (about_window), 10);

      dialog_vbox1 = GTK_DIALOG (about_window)->vbox;
      gtk_object_set_data (GTK_OBJECT (about_window), "dialog_vbox1", dialog_vbox1);
      gtk_widget_show (dialog_vbox1);
      gtk_container_border_width (GTK_CONTAINER (dialog_vbox1), 5);

      hbox1 = gtk_hbox_new (FALSE, 0);
      gtk_object_set_data (GTK_OBJECT(about_window), "hbox1", hbox1);
      gtk_widget_show (hbox1);
      gtk_box_pack_start (GTK_BOX(dialog_vbox1), hbox1, TRUE, TRUE, 0);
      gtk_container_border_width (GTK_CONTAINER(hbox1), 5);
      gtk_widget_realize (about_window);

      label1 = gtk_label_new (DESCRIPTION " v" VERSION " by Sam Clegg\nhttp://www.superduper.net/xmms-mad/\nBased on libmad v" MAD_VERSION " by Rob Leslie\nhttp://mad.sourceforge.net");
      gtk_object_set_data (GTK_OBJECT (about_window), "label1", label1);
      gtk_widget_show (label1);
      gtk_box_pack_start (GTK_BOX (hbox1), label1, TRUE, TRUE, 0);

      dialog_action_area1 = GTK_DIALOG (about_window)->action_area;
      gtk_object_set_data (GTK_OBJECT (about_window), "dialog_action_area1", dialog_action_area1);
      gtk_widget_show (dialog_action_area1);
      gtk_container_border_width (GTK_CONTAINER (dialog_action_area1), 10);

      about_exit = gtk_button_new_with_label ("Ok");
      gtk_signal_connect_object (GTK_OBJECT (about_exit), "clicked",
                                 GTK_SIGNAL_FUNC (gtk_widget_destroy),
                                 GTK_OBJECT (about_window));

      gtk_object_set_data (GTK_OBJECT (about_window), "about_exit", about_exit);
      gtk_widget_show (about_exit);
      gtk_box_pack_start (GTK_BOX (dialog_action_area1), about_exit, TRUE, TRUE, 0);

      gtk_widget_show (about_window);
    }
  else
    {
      gdk_window_raise (about_window->window);
    }
}

/**
 * Display a GTK box containing the given error message.
 * Taken from mpg123 plugin.
 */
void
xmmsmad_error (char *error, ...)
{
  if (!error_dialog)
    {
      va_list args;
      char string[256];
      va_start (args, error);
      vsnprintf (string, 256, error, args);
      va_end (args);
      GDK_THREADS_ENTER ();
      error_dialog = xmms_show_message ("Error", string, "Ok", FALSE, 0, 0);
      gtk_signal_connect (GTK_OBJECT (error_dialog),
                          "destroy",
                          GTK_SIGNAL_FUNC (gtk_widget_destroyed),
                          &error_dialog);
      GDK_THREADS_LEAVE ();
    }
}

extern void xmmsmad_get_file_info (char *filename);

/**
 * XMMS input plugin hook.
 */
InputPlugin *
get_iplugin_info (void)
{
  memset (&mad_plugin, 0, sizeof (InputPlugin));
  mad_plugin.description   = DESCRIPTION " " VERSION;
  mad_plugin.about         = xmmsmad_about;
  mad_plugin.init          = xmmsmad_init;
  mad_plugin.set_eq        = xmmsmad_set_eq;
  mad_plugin.configure     = xmmsmad_configure;
  mad_plugin.is_our_file   = xmmsmad_is_our_file;
  mad_plugin.play_file     = xmmsmad_play_file;
  mad_plugin.stop          = xmmsmad_stop;
  mad_plugin.pause         = xmmsmad_pause;
  mad_plugin.seek          = xmmsmad_seek;
  mad_plugin.get_time      = xmmsmad_get_time;
  mad_plugin.get_song_info = xmmsmad_get_song_info;
  mad_plugin.file_info_box = xmmsmad_get_file_info;
  mad_plugin.cleanup       = xmmsmad_cleanup;

  return &mad_plugin;
}
